/*
SPHEREC.C - Client RPC sphere demo program - 1994 James M. Finnegan
*/
#include <windows.h>
#include <stdlib.h>
#include <time.h>
#include "sphere.h"    // Header file generated by the MIDL compiler
#include "resource.h"


char gszServer[80];	    // Server name
char gszProtocol[80];	// Protocol sequence
char gszEndpoint[80];	// Endpoint

int giMaxX,giMaxY,      // max window client size
    giCurX,giCurY;      // current window client size
           
HDC     ghHiddenDC;     // DC for hidden bitmap
HBITMAP ghFrame;        // handle for hidden bitmap
HBITMAP ghPrevFrame;  	// placeholder for deselected bitmap
HBRUSH  ghBrush[2];     // Handles for white/red brush

short giBelts=18;		// Number of belts to draw
BOOL  bPoints=FALSE,
      bBuffered=TRUE;

HINSTANCE ghInst;

                                              
LRESULT WINAPI WndProc(HWND,UINT,WPARAM,LPARAM);
LONG WINAPI ServerProc(HWND,UINT,WPARAM,LPARAM);
LONG WINAPI ConfigDlgProc(HWND,UINT,WPARAM,LPARAM);
RPC_STATUS BindToServer(HWND);
void TimerProc(HDC);
void DrawSphere(HDC,short);
HDC  CreateHiddenBitmap(HDC);
void ClearDC(HDC);


int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                                            LPSTR lpszCmdLine, int nCmdShow)
{
    static char szAppName[]="RPC Sphere!";
    HWND        hWnd;
    MSG         msg;
    WNDCLASS    wndclass;

    
    ghInst=hInstance;

    // Get the maximum possible bitmap size
    giMaxX=GetSystemMetrics(SM_CXFULLSCREEN);
    giMaxY=GetSystemMetrics(SM_CYFULLSCREEN);
        
    if(!hPrevInstance)
    {
        wndclass.style         = CS_HREDRAW | CS_VREDRAW;
        wndclass.lpfnWndProc   = WndProc;
        wndclass.cbClsExtra    = 0;
        wndclass.cbWndExtra    = DLGWINDOWEXTRA;
        wndclass.hInstance     = hInstance;
        wndclass.hIcon         = NULL;
        wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wndclass.hbrBackground = GetStockObject(BLACK_BRUSH);
        wndclass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
        wndclass.lpszClassName = szAppName;

        if(!RegisterClass(&wndclass))
            return -1;
    }

    // Create the main window
    hWnd= CreateWindow(szAppName,szAppName,
                WS_OVERLAPPED | WS_THICKFRAME | WS_MINIMIZEBOX | 
                WS_MAXIMIZEBOX |WS_SYSMENU | WS_CLIPCHILDREN,
                CW_USEDEFAULT,CW_USEDEFAULT,
                CW_USEDEFAULT,CW_USEDEFAULT,
                0,0,hInstance,NULL);

    
    ShowWindow(hWnd,nCmdShow);
    UpdateWindow(hWnd);
    
    while(GetMessage(&msg,NULL,0,0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return msg.wParam;
}                 


LRESULT WINAPI WndProc(HWND hWnd,UINT uMessage,WPARAM wParam,LPARAM lParam)
{
    static HDC hDC;


    switch(uMessage)
    {
        case WM_CREATE:
            // Query for a server name
            DialogBox(ghInst,MAKEINTRESOURCE(IDD_DIALOG2),hWnd,ServerProc);
            // Bind to the server
            BindToServer(hWnd);
            
			// Get the window's DC, hang onto it until the app terminates...
            hDC=GetDC(hWnd);
			// Create a hidden bitmap (for buffering)
            CreateHiddenBitmap(hDC);
			// Kick it off!
            SetTimer(hWnd,1,50,NULL);
            break;

        case WM_TIMER:
            // Paint a sphere
            TimerProc(hDC);
            break;
            
        case WM_SIZE:
            // Remember the current client size
            giCurX = LOWORD(lParam);
            giCurY = HIWORD(lParam);

            // Set mapping mode and extents of the client area
            SetMapMode(hDC,MM_ISOTROPIC);
            SetWindowExtEx(hDC,200,200,NULL); // Set Logical extents
            SetViewportExtEx(hDC,giCurX / 2, -giCurY / 2,NULL);
            SetViewportOrgEx(hDC,giCurX / 2,  giCurY / 2,NULL);
            
            // Set mapping mode and extents of the hidden bitmap
            SetMapMode(ghHiddenDC,MM_ISOTROPIC);
            SetWindowExtEx(ghHiddenDC,200,200,NULL); // Set Logical extents
            SetViewportExtEx(ghHiddenDC,giCurX / 2, -giCurY / 2,NULL);
            SetViewportOrgEx(ghHiddenDC,giCurX / 2,  giCurY / 2,NULL);
            break;

        case WM_COMMAND:
            switch(LOWORD(wParam)) 
            {
                case ID_FILE_EXIT:
                    PostMessage(hWnd,WM_CLOSE,0,0);
                    break;
                
                case ID_OPTIONS_CONFIGURE:
                    DialogBox(ghInst,MAKEINTRESOURCE(IDD_DIALOG1),hWnd,ConfigDlgProc);
                    break;

                default:
                    break;

            }
            break;

        case WM_DESTROY:
            // Delete the bitmap, brushes, and DC
            SelectObject(ghHiddenDC,ghPrevFrame);
            DeleteObject(ghFrame);
            DeleteDC(ghHiddenDC);

            DeleteObject(ghBrush[0]);
            DeleteObject(ghBrush[1]);

            ReleaseDC(hWnd,hDC);
    
            PostQuitMessage(0);
            break;
        
        default:
            return DefWindowProc(hWnd,uMessage,wParam,lParam);
            break;
          
    }
    return 0L;
}                                                            
       

/*
ServerProc - Dialog box procedure for the RPC server binding.  Establishes
             global variables for the server, protocol, and endpoint.
*/
LRESULT WINAPI ServerProc(HWND hWnd,UINT uMessage,WPARAM wParam,LPARAM lParam)
{
    switch(uMessage)
    {
        case WM_INITDIALOG:
            SendMessage(GetDlgItem(hWnd,IDC_LPC),BM_SETCHECK,1,0);
            break;
        
        case WM_CLOSE:
            PostMessage(hWnd,WM_COMMAND,IDCANCEL,0L);
            break;
            
        case WM_COMMAND:
            switch(wParam)
            {
                case IDOK:
                    // If user selected LPC's
                    if(SendMessage(GetDlgItem(hWnd,IDC_LPC),BM_GETCHECK,0,0))
                    {
                        lstrcpy(gszProtocol,"ncalrpc");
                        lstrcpy(gszEndpoint,"spheresrv");
                        lstrcpy(gszServer,"");
                    }
                    else
                    {
                        GetDlgItemText(hWnd, IDC_EDIT1, gszServer, 79);
                        lstrcpy(gszProtocol,"ncacn_nb_nb");
                        lstrcpy(gszEndpoint,"123");
                    }

                    EndDialog(hWnd,TRUE);
                    break;

                case IDCANCEL:
                    EndDialog(hWnd,FALSE);
                    break;
            }
            break;
            
        default:
            return FALSE;
    }
    return TRUE;
}


/*
ConfigDlgProc - Dialog box procedure for the sphere configuration.  Establishes
                global attribute variables for the sphere
*/
LRESULT WINAPI ConfigDlgProc(HWND hWnd,UINT uMessage,WPARAM wParam,LPARAM lParam)
{
    BOOL bTest;

    switch(uMessage)
    {
        case WM_INITDIALOG:
            SetDlgItemInt(hWnd,IDC_BELTS,giBelts,FALSE);
            SendMessage(GetDlgItem(hWnd,IDC_POINTS),BM_SETCHECK,bPoints,0);
            SendMessage(GetDlgItem(hWnd,IDC_CHECKERED),BM_SETCHECK,!bPoints,0);

            SendMessage(GetDlgItem(hWnd,IDC_BUFFERED),BM_SETCHECK,bBuffered,0);
            SendMessage(GetDlgItem(hWnd,IDC_DIRECT),  BM_SETCHECK,!bBuffered,0);
            break;
            
        case WM_CLOSE:
            PostMessage(hWnd,WM_COMMAND,IDCANCEL,0L);
            break;
            
        case WM_COMMAND:
            switch(wParam)
            {
                case IDOK:
                    giBelts=GetDlgItemInt(hWnd,IDC_BELTS,&bTest,FALSE);
                    // Limit the number of belts
                    if(giBelts < 2)
                        giBelts=2;
                    if(giBelts > 35)
                        giBelts = 35;

                    bPoints=SendMessage(GetDlgItem(hWnd,IDC_POINTS),BM_GETCHECK,0,0);
                    bBuffered=SendMessage(GetDlgItem(hWnd,IDC_BUFFERED),BM_GETCHECK,0,0);

                    EndDialog(hWnd,TRUE);
                    break;

                case IDCANCEL:
                    EndDialog(hWnd,FALSE);
                    break;
            }
            break;
            
        default:
            return FALSE;
    }
    return TRUE;
}
             

/*
BindToServer - Bind to the server using the global variables for the protocol
			   sequence, server name, and endpoint (established in ServerProc)
*/
RPC_STATUS BindToServer(HWND hWnd)
{
    RPC_STATUS iStatus;
    unsigned char * pszStringBinding   = NULL;
    char pszText[80];


    // Compose a string binding
    iStatus = RpcStringBindingCompose(NULL,
                                     gszProtocol,
                                     gszServer, 
                                     gszEndpoint,
                                     NULL,
                                     &pszStringBinding);
    if(iStatus) 
    {
        wsprintf(pszText,"RpcStringBindingCompose failed: %d\n",iStatus);
        MessageBox(hWnd,pszText,"RPC Runtime Error",MB_ICONEXCLAMATION);
        return iStatus;
    }

    // Bind using the created string binding...
    iStatus = RpcBindingFromStringBinding(pszStringBinding,&hSphere);
    if(iStatus) 
    {
        wsprintf(pszText,"RpcBindingFromStringBinding failed: %d\n\n",iStatus);
        MessageBox(hWnd,pszText,"RPC Runtime Error",MB_ICONEXCLAMATION);
        return iStatus;
    }

    return iStatus;
}


/*
TimerProc - Processes timer message for the main window procedure
*/
void TimerProc(HDC hDC)
{
    static BOOL bInTimer=FALSE;


    // If we are already processing a timer message, bail...
    if(bInTimer==TRUE)
        return;

    bInTimer=TRUE;

    // If the user wants to buffer the drawing...
    if(bBuffered)
    {
        // clear the bitmap
        ClearDC(ghHiddenDC);
        // Draw the aphere onto the bitmap    
        DrawSphere(ghHiddenDC,giBelts);
		// copy it to the display DC
        BitBlt(hDC,-200,200,400,-400,ghHiddenDC,-200,200,SRCCOPY);
    }
    // Otherwise, draw directly to the window
    else
    {
        // Clear the window
        ClearDC(hDC);
        // Draw it!
        DrawSphere(hDC,giBelts);
    }

    bInTimer=FALSE;
}


/*
DrawSphere - Renders a sphere of iNumBelts using RPC, and paints 
             the results on the passed in hDC
*/
void DrawSphere(HDC hDC,short iNumBelts)
{
    char szText[80];
    short int  iPointX[MAX_BELTS][MAX_SIDES],
               iPointY[MAX_BELTS][MAX_SIDES];
    short int  bVisible[MAX_BELTS][MAX_SIDES];

    short i,j;
    
    POINT point[4];
    
    static short int iTiltX=-1,  // Initial tilt
                     iTiltY=0,
                     iTiltZ=0;

    // Set up a random initial tilt...
    if(iTiltX == -1)
    {
        srand((unsigned)time(NULL));
        
        iTiltX=(short)rand()%360;  // Initial tilt
        iTiltY=(short)rand()%360;
        iTiltZ=(short)rand()%360;
    }
    // Increment tilt (in degrees)
    if((iTiltX=iTiltX+1) > 359)
        iTiltX=0;
    if((iTiltY=iTiltY+1) > 359)
        iTiltY=0;
    if((iTiltZ=iTiltZ+1) > 359)
        iTiltZ=0;
    
    // Render the sphere on the server
    RpcTryExcept
    {
        RPCRenderSphere(iNumBelts,iTiltX,iTiltY,iTiltZ,iPointX,iPointY,bVisible);
    }
    RpcExcept(1)
    {
        // Tell the user that things went amuck
        wsprintf(szText,"Exception %ld",RpcExceptionCode());
        MessageBox(NULL,szText,"RPC Exception",MB_OK);
    }
    RpcEndExcept;

    // select the current belt
    for(i=0;i<giBelts;i++)
    {
        // Close the belt by pointing the last element to the first
        iPointX[i][giBelts*2+1]=iPointX[i][0];
        iPointY[i][giBelts*2+1]=iPointY[i][0];
        
        // Draw each side of the current belt
        for(j=0;j<giBelts*2;j++)
        {
            if(bPoints)
            {    
                // If the point is facing away from the viewer, draw it blue
                if(bVisible[i][j] == FALSE)
                    SetPixel(hDC,iPointX[i][j],iPointY[i][j],RGB(0,0,255));
                else
                    SetPixel(hDC,iPointX[i][j],iPointY[i][j],RGB(255,255,255));
            }
            else
            {
                // If the current point is visible, draw a polygon
                if(bVisible[i][j] != FALSE)
                {
                    // Get first set of points from the current belt
                    point[0].x = iPointX[i][j+1];
                    point[0].y = iPointY[i][j+1];
                    point[1].x = iPointX[i][j];
                    point[1].y = iPointY[i][j];
    
                    // Get next set of points from the next belt
                    point[2].x = iPointX[i+1][j];
                    point[2].y = iPointY[i+1][j];
                    point[3].x = iPointX[i+1][j+1];
                    point[3].y = iPointY[i+1][j+1];
                    
                    // Select the brush out of the array, alternating
                    // red and white, depending on the belt/side
                    SelectObject(hDC,ghBrush[(i%2)+(j%2) &1]);

                    // Draw it!
                    Polygon(hDC,point,4);
                }    
            }    
        }
    }
}


/*
CreateHiddenBitmap - Creates a memory DC compatible with the passed in DC
*/
HDC CreateHiddenBitmap(HDC hDC)
{
    ghHiddenDC= CreateCompatibleDC(hDC);
    ghFrame=    CreateCompatibleBitmap(hDC,giMaxX,giMaxY);
    if(ghFrame==NULL)
    {
        DeleteDC(ghHiddenDC);
        return NULL;
    }
    
    ghBrush[0]=CreateSolidBrush(RGB(255,0,0));     // Red brush
    ghBrush[1]=CreateSolidBrush(RGB(255,255,255)); // White brush

    ghPrevFrame= SelectObject(ghHiddenDC,ghFrame);
    ClearDC(ghHiddenDC);   

    return ghHiddenDC;
}


/*
ClearDC - Fills the passed DC with a black brush
*/
void ClearDC(HDC hDC)
{
    HRGN hRgn;
    HBRUSH ghBrush;


    // Select the entire logical range
    hRgn = CreateRectRgn(-1000,-1000,1000,1000);
    SelectObject(hDC, hRgn);

    ghBrush = CreateSolidBrush(RGB(0,0,0));

    // Fill the range with the brush
    FillRgn(hDC, hRgn, ghBrush);

    DeleteObject(hRgn);
    DeleteObject(ghBrush);
}


/*
midl_user_allocate/midl_user_free - RPC required RPC allocation/deallocation routines
*/
void __RPC_FAR * __RPC_API midl_user_allocate(size_t len)
{
    return(malloc(len));
}

void __RPC_API midl_user_free(void __RPC_FAR * ptr)
{
    free(ptr);
}


