/*
SPHERES.C - Server RPC sphere demo program - 1994 James M. Finnegan
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "sphere.h"    // Header file generated by the MIDL compiler

void   RotateBelt(double, double, double, int, double [][3], double [][3], int, int [], int []);
int    IsVisible(double, double, double, double, double, double, double, double, double);
double DegreesToRadians(double);


RPC_STATUS main()
{
    RPC_STATUS lStatus;


    printf("RPC Sphere server demo application - 1994 James M. Finnegan\n\n");

    printf("Using LPC with endpoint 'spheresrv'\n");
    lStatus=RpcServerUseProtseqEp("ncalrpc",20,"spheresrv",NULL);
                                   
    if(lStatus) 
    {
        printf("LPC protocol sequence failed!\n");
        return lStatus;
    }

    printf("Using NetBIOS over NetBEUI with endpoint '123'\n");
    lStatus=RpcServerUseProtseqEp("ncacn_nb_nb",20,"123",NULL);
                                   
    if(lStatus) 
    {
        printf("NetBIOS protocol sequence failed!\n");
        return lStatus;
    }

    printf("Registering server interface\n");
    lStatus = RpcServerRegisterIf(sphere_ServerIfHandle, // interface to register
                                 NULL,                   // MgrTypeUuid
                                 NULL);                  // MgrEpv; null means use default

    if(lStatus) 
    {
        printf("Interface registration failed!\n");
        return lStatus;
    }

    printf("Listening for client applications...\n");
    lStatus = RpcServerListen(1,40,FALSE);
    
    printf("RpcServerListen returned %ld\n", lStatus);
    if(lStatus) 
    {
        return lStatus;
    }

    return 0;
}


/*
RPCRenderSphere - Renders a sphere within a logical x/y/z coordinate
                  system of 400 units.  Draws iNumBelts*2 points along
                  each belt (iNumBelts).  x/y screen coordinates are 
                  returned in an array in iPointX/iPointY.  Point
                  visibility is returned in bVisible.
*/
void RPCRenderSphere(short iNumBelts,short iTiltX,short iTiltY,short iTiltZ,
                     short iPointX[][MAX_SIDES],short iPointY[][MAX_SIDES],
                                               short bVisible[][MAX_SIDES])
{
    int vx[MAX_SIDES], vy[MAX_SIDES];

    double dYaw, dRoll, dPitch;

    int     iSidesPerBelt = iNumBelts * 2;
    double  dLongitudeScale = 3.141592654 / (iSidesPerBelt/2);
    double  dLatitudeScale  = 3.141592654 / (iNumBelts);  

    double  dLatitude,
            dLongitude;

    double  dCurrentRotatedPoints [MAX_SIDES][3], 
            dPreviousRotatedPoints[MAX_SIDES][3], 
            dPoints [MAX_SIDES][3];
    
    int i,j,iCurrentBelt,iRadius=3;
    int iDistance=900;   // Viewers distance from viewplane

    

    printf("Calculating a %d belt sphere  X=%3d Y=%3d Z=%3d\n",iNumBelts,
                                                     iTiltX,iTiltY,iTiltZ);

    // convert the passed in angles to radians
    dYaw   = DegreesToRadians(iTiltX);   // rotation
    dRoll  = DegreesToRadians(iTiltY);   // axis tilt
    dPitch = DegreesToRadians(iTiltZ);   // tilt to or from viewer
    
    // Initialize points to the sphere's pole
    for(i=0;i<=iSidesPerBelt+1;i++)
    {
        dPoints[i][0] = (double)0;
        dPoints[i][1] = (double)iRadius;
        dPoints[i][2] = (double)0;
    }

    // Rotate the belt, passing points in dPoints,
    // storing rotated points in dCurrentRotatedPoints, and 
    // screen coordinates in vx/vy
    RotateBelt(dYaw, dRoll, dPitch, iSidesPerBelt+1,dCurrentRotatedPoints,dPoints,iDistance,vx,vy);

    dLongitude = dLongitudeScale;
    // Paint middle
    for(iCurrentBelt=0; iCurrentBelt < iNumBelts; iCurrentBelt++)
    {
        // Copy data into 'previous' buffer
        for(i=0; i < iSidesPerBelt + 1; i++)
        {
            for(j=0; j<3; j++)
                dPreviousRotatedPoints[i][j] = dCurrentRotatedPoints[i][j];
            
            iPointX[iCurrentBelt][i] = vx[i];
            iPointY[iCurrentBelt][i] = vy[i];
        }

        // Calculate points around the sphere
        for(dLatitude=0,i=0; i < iSidesPerBelt; i++,dLatitude += dLatitudeScale)
        {
            dPoints[i][0] = cos(dLatitude)  * sin(dLongitude) * iRadius;
            dPoints[i][1] = cos(dLongitude) * iRadius;
            dPoints[i][2] = sin(dLatitude)  * sin(dLongitude) * iRadius;
        }

        // Close the belt!
        dPoints[iSidesPerBelt][0]=dPoints[0][0];
        dPoints[iSidesPerBelt][1]=dPoints[0][1];
        dPoints[iSidesPerBelt][2]=dPoints[0][2];

        // Rotate the belt, passing points in dPoints,
        // storing rotated points in dCurrentRotatedPoints, and 
        // screen coordinates in vx/vy
        RotateBelt(dYaw, dRoll, dPitch, iSidesPerBelt + 1, dCurrentRotatedPoints, dPoints, iDistance, vx, vy);

        for(i=0; i < iSidesPerBelt; i++)
        {
            // Set visibility
            if(iCurrentBelt == 0)
            {
                bVisible[iCurrentBelt][i]=
                   IsVisible(dCurrentRotatedPoints [i+1][0],dCurrentRotatedPoints [i+1][1],dCurrentRotatedPoints [i+1][2],
                             dCurrentRotatedPoints [i]  [0],dCurrentRotatedPoints [i]  [1],dCurrentRotatedPoints [i]  [2],
                             dPreviousRotatedPoints[0]  [0],dPreviousRotatedPoints[0]  [1],dPreviousRotatedPoints[0]  [2]);
            }
            else
            {
                bVisible[iCurrentBelt][i]=
                   IsVisible(dPreviousRotatedPoints[i]  [0],dPreviousRotatedPoints[i]  [1],dPreviousRotatedPoints[i]  [2],
                             dPreviousRotatedPoints[i+1][0],dPreviousRotatedPoints[i+1][1],dPreviousRotatedPoints[i+1][2], 
                             dCurrentRotatedPoints [i+1][0],dCurrentRotatedPoints [i+1][1],dCurrentRotatedPoints [i+1][2]);
            }
        }                
        
        dLongitude += dLongitudeScale;
    }

    // copy the pole (since we dropped out of the for loop)
    for(i=0; i < iSidesPerBelt + 1; i++)
    {
        iPointX[iCurrentBelt][i] = vx[i];
        iPointY[iCurrentBelt][i] = vy[i];
    }
}
    

/*
RotateBelt - Rotates and translates a series of points into xyz and xy 
             coordinates
*/
void RotateBelt( double dYaw, double dRoll, double dPitch,
                 int iPoints, double rot_pts[][3], double pt[][3],
                 int iDistance, int vx[], int vy[])
{
    int i;
    double xa,ya,za,x,y,z;


    // Do the entire belt...
    for(i=0; i < iPoints; i++)
    {
        // Use matrix math to rotate the points...
        xa =  - cos(dYaw) * pt[i][0] - sin(dYaw)   * pt[i][2];
        za =  - sin(dYaw) * pt[i][0] + cos(dYaw)   * pt[i][2];
        x  =  cos(dRoll)  * xa       + sin(dRoll)  * pt[i][1];
        ya =  cos(dRoll)  * pt[i][1] - sin(dRoll)  * xa;
        z  =  cos(dPitch) * za       - sin(dPitch) * ya;
        y  =  sin(dPitch) * za       + cos(dPitch) * ya;
        
        // After rotation, move it back on the z axis
        z -= 15;
        
        // Assign x/y/z array of rotated points
        rot_pts[i][0] = x;
        rot_pts[i][1] = y;
        rot_pts[i][2] = z;

        // Calculate x/y coordinates on the viewplane
        vx[i] = (int)( iDistance * x / z);
        vy[i] = (int)(-iDistance * y / z);
    }
}


/*
IsVisible -  Tests a polygon's visibility by using backface culling, or the
             'plane equation method'

Returns: True if visible
*/
int IsVisible(double x1, double y1, double z1, 
              double x2, double y2, double z2,
              double x3, double y3, double z3) 
{
    double dVisible;


    // Use backface culling (plane equation method) to calculate visibility
    // in relation to the viewer
    dVisible = x1*(y3*z2 - y2*z3) - x2*(y3*z1 - y1*z3) - x3*(y1*z2 - y2*z1);

    // If it's visible...
    if(dVisible < 0)
        return TRUE;
    else
        return FALSE;
}


/*
DegreesToRadians - Converts dDegrees into radians by multiplying dDegrees
                   by (PI / 180)
*/
double DegreesToRadians(double dDegrees)
{
    double dRadiansPerDegree=0.017453292;  // PI / 180
    
    
    // Adjust the passed paramter to a positive angle
    while(dDegrees >= 360)
        dDegrees -= 360;
    while(dDegrees < 0)
        dDegrees += 360;
    
    return dRadiansPerDegree * dDegrees;
}


/*
midl_user_allocate/midl_user_free - RPC required RPC allocation/deallocation routines
*/
void __RPC_FAR * __RPC_API midl_user_allocate(size_t len)
{
    return(malloc(len));
}


void __RPC_API midl_user_free(void __RPC_FAR * ptr)
{
    free(ptr);
}
